#!/usr/bin/python
# -*- coding: utf-8 -*-

"""Module implements a Python wrapper for the fingerprint extractor 'audioid-mon-extract' of audioid monitor generation 3."""

__author__    = "Lars Fabig"
__date__      = "14.08.2016"
__email__     = "lfabig@mufin.com"
__docformat__ = 'reStructuredText'


# Standard library imports ...
import pkg_resources
import shutil
import urllib2
import urlparse

# Third-party imports ...
import requests

# Local imports ...
import exceptions


class AudioidMonExtractorService:
    """Class implements the Python wrapper for 'audioid-mon-extract'."""

    def __init__(self, host_address, port = "80"):
        """Init class with host address."""
        hostname = "http://" + urlparse.urlparse(host_address).hostname + ":" + str(port)
        self.__base_url = urllib2.urlparse.urljoin(hostname, "audioid-mon/v1")


    def get_base_url(self):
        """Return base url"""
        return self.__base_url


    def get_client_version(self):
        """Return version of package"""
        return pkg_resources.get_distribution('pyaudioidmon').version


    def extract_fingerprint(self, audio_filename, fingerprint_filename):
        """Extract a fingerprint from an audio file and return a fingerprint file.

        :param audio_filename: Name of the audio file
        :type audio_filename: string
        :param fingerprint_filename: Name of fingerprint file
        :type fingerprint_filename: string
        :raises: :exc:`AudioidMonInvalidInput` if the fingerprint does not exists.
        :raises: :exc:`AudioidMonSubmittedItemIsTooLarge` if the audio file is too large.
        :raises: :exc:`AudioidMonUnexpectedError` if an expected error occurs.
        :raises: :exc:`AudioidMonServiceNotAvailable` if a connection error occurs.

        :Example:

        >>> import audioidmon
        >>> client = audioidmon.AudioidMonExtractorService("http://127.0.0.1")
        >>> client.extract_fingerprint("advertisement.mp3", "advertisement.mfp")

        """
        try:
            url = self.__base_url+"/extract/sync"
            files = {"upload": open(audio_filename, "rb")}
            resp = requests.post(url, files=files)
            try:
                resp.raise_for_status()
                with open(fingerprint_filename, 'wb') as file:
                    for chunk in resp:
                        file.write(chunk)
                return resp
            except requests.HTTPError as err:
                try:
                    if resp.json()["code"] == 20109:
                        raise exceptions.AudioidMonInvalidInput(repr(resp))
                    elif resp.json()["code"] == 20114:
                        raise exceptions.AudioidMonSubmittedItemIsTooLarge(repr(resp))
                    elif resp.json()["code"] == 20109:
                        raise exceptions.AudioidMonInvalidInput(repr(resp))
                    else:
                        raise exceptions.AudioidMonUnexpectedError(repr(resp))
                except (ValueError, KeyError):
                    raise exceptions.AudioidMonUnexpectedError(err)
        except requests.ConnectionError as err:
            raise exceptions.AudioidMonServiceNotAvailable(err)
        except requests.RequestException as err:
            raise exceptions.AudioidMonUnexpectedError(err)


if __name__ == "__main__":
    import doctest
    doctest.testmod()
